@@ -53,6 +53,7 @@ pub fn wait_timeout(child: &mut Child, dur: Duration)
 }
 
 impl State {
+    #[allow(unused_assignments)]
     fn init() {
         unsafe {
             // Create our "self pipe" and then set both ends to nonblocking
@@ -69,7 +70,28 @@ impl State {
             // Register our sigchld handler
             let mut new: libc::sigaction = mem::zeroed();
             new.sa_sigaction = sigchld_handler as usize;
-            new.sa_flags = libc::SA_NOCLDSTOP | libc::SA_RESTART;
+
+            // FIXME: remove this workaround when the PR to libc get merged and released
+            //
+            // This is a workaround for the type mismatch in the definition of SA_*
+            // constants for android. See https://github.com/rust-lang/libc/pull/511
+            //
+            // Do $value as type_of($target)
+            macro_rules! _as {
+                ($value:expr, $target:expr) => (
+                    {
+                        let mut x = $target;
+                        x = $value as _;
+                        x
+                    }
+                )
+            }
+
+            let sa_flags = new.sa_flags;
+            new.sa_flags = _as!(libc::SA_NOCLDSTOP, sa_flags) |
+                           _as!(libc::SA_RESTART, sa_flags) |
+                           _as!(libc::SA_SIGINFO, sa_flags);
+
             assert_eq!(libc::sigaction(libc::SIGCHLD, &new, &mut state.prev), 0);
 
             STATE = mem::transmute(state);
@@ -264,9 +286,28 @@ fn now_ns() -> u64 {
 // it. At that point we're guaranteed that there's something in the pipe
 // which will wake up the other end at some point, so we just allow this
 // signal to be coalesced with the pending signals on the pipe.
-extern fn sigchld_handler(_signum: c_int) {
-    let state = unsafe { &*STATE };
-    notify(&state.write);
+extern fn sigchld_handler(signum: c_int,
+                          info: *mut libc::siginfo_t,
+                          ptr: *mut libc::c_void) {
+    type FnSigaction = extern fn(c_int, *mut libc::siginfo_t, *mut libc::c_void);
+    type FnHandler = extern fn(c_int);
+
+    unsafe {
+        let state = &*STATE;
+        notify(&state.write);
+
+        let fnptr = state.prev.sa_sigaction;
+        if fnptr == 0 {
+            return
+        }
+        if state.prev.sa_flags & libc::SA_SIGINFO == 0 {
+            let action = mem::transmute::<usize, FnHandler>(fnptr);
+            action(signum)
+        } else {
+            let action = mem::transmute::<usize, FnSigaction>(fnptr);
+            action(signum, info, ptr)
+        }
+    }
 }
 
 impl ExitStatus {
